import range from 'lodash.range';
import React from 'react';
import PropTypes from 'prop-types';
import classNames from 'classnames';
import difference from 'lodash.difference';
import find from 'lodash.find';

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _nonIterableRest();
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArrayLimit(arr, i) {
  var _arr = [];
  var _n = true;
  var _d = false;
  var _e = undefined;

  try {
    for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance");
}

const SORTED_PITCHES = ['C', 'Db', 'D', 'Eb', 'E', 'F', 'Gb', 'G', 'Ab', 'A', 'Bb', 'B'];
const ACCIDENTAL_PITCHES = ['Db', 'Eb', 'Gb', 'Ab', 'Bb'];
const PITCH_INDEXES = {
  C: 0,
  'C#': 1,
  Db: 1,
  D: 2,
  'D#': 3,
  Eb: 3,
  E: 4,
  F: 5,
  'F#': 6,
  Gb: 6,
  G: 7,
  'G#': 8,
  Ab: 8,
  A: 9,
  'A#': 10,
  Bb: 10,
  B: 11
};
const MIDI_NUMBER_C0 = 12;
const MIN_MIDI_NUMBER = MIDI_NUMBER_C0;
const MAX_MIDI_NUMBER = 127;
const NOTE_REGEX = /([a-g])([#b]?)(\d+)/;
const NOTES_IN_OCTAVE = 12; // Converts string notes in scientific pitch notation to a MIDI number, or null.
//
// Example: "c#0" => 13, "eb5" => 75, "abc" => null
//
// References:
// - http://www.flutopedia.com/octave_notation.htm
// - https://github.com/danigb/tonal/blob/master/packages/note/index.js

function fromNote(note) {
  if (!note) {
    throw Error('Invalid note argument');
  }

  const match = NOTE_REGEX.exec(note.toLowerCase());

  if (!match) {
    throw Error('Invalid note argument');
  }

  const _match = _slicedToArray(match, 4),
        letter = _match[1],
        accidental = _match[2],
        octave = _match[3];

  const pitchName = `${letter.toUpperCase()}${accidental}`;
  const pitchIndex = PITCH_INDEXES[pitchName];

  if (pitchIndex == null) {
    throw Error('Invalid note argument');
  }

  return MIDI_NUMBER_C0 + pitchIndex + NOTES_IN_OCTAVE * parseInt(octave, 10);
} //
// Build cache for getAttributes
//


function buildMidiNumberAttributes(midiNumber) {
  const pitchIndex = (midiNumber - MIDI_NUMBER_C0) % NOTES_IN_OCTAVE;
  const octave = Math.floor((midiNumber - MIDI_NUMBER_C0) / NOTES_IN_OCTAVE);
  const pitchName = SORTED_PITCHES[pitchIndex];
  return {
    note: `${pitchName}${octave}`,
    pitchName,
    octave,
    midiNumber,
    isAccidental: ACCIDENTAL_PITCHES.includes(pitchName)
  };
}

function buildMidiNumberAttributesCache() {
  return range(MIN_MIDI_NUMBER, MAX_MIDI_NUMBER + 1).reduce((cache, midiNumber) => {
    cache[midiNumber] = buildMidiNumberAttributes(midiNumber);
    return cache;
  }, {});
}

const midiNumberAttributesCache = buildMidiNumberAttributesCache(); // Returns an object containing various attributes for a given MIDI number.
// Throws error for invalid midiNumbers.

function getAttributes(midiNumber) {
  const attrs = midiNumberAttributesCache[midiNumber];

  if (!attrs) {
    throw Error('Invalid MIDI number');
  }

  return attrs;
} // Returns all MIDI numbers corresponding to natural notes, e.g. C and not C# or Bb.


const NATURAL_MIDI_NUMBERS = range(MIN_MIDI_NUMBER, MAX_MIDI_NUMBER + 1).filter(midiNumber => !getAttributes(midiNumber).isAccidental);
var MidiNumbers = {
  fromNote,
  getAttributes,
  MIN_MIDI_NUMBER,
  MAX_MIDI_NUMBER,
  NATURAL_MIDI_NUMBERS
};

class Key extends React.Component {
  constructor(...args) {
    super(...args);

    _defineProperty(this, "playNote", () => {
      this.props.onPlayNote(this.props.midiNumber);
    });

    _defineProperty(this, "stopNote", () => {
      this.props.onStopNote(this.props.midiNumber);
    });
  }

  // Key position is represented by the number of natural key widths from the left
  getAbsoluteKeyPosition(midiNumber) {
    const OCTAVE_WIDTH = 7;

    const _MidiNumbers$getAttri = MidiNumbers.getAttributes(midiNumber),
          octave = _MidiNumbers$getAttri.octave,
          pitchName = _MidiNumbers$getAttri.pitchName;

    const pitchPosition = this.props.pitchPositions[pitchName];
    const octavePosition = OCTAVE_WIDTH * octave;
    return pitchPosition + octavePosition;
  }

  getRelativeKeyPosition(midiNumber) {
    return this.getAbsoluteKeyPosition(midiNumber) - this.getAbsoluteKeyPosition(this.props.noteRange.first);
  }

  render() {
    const _this$props = this.props,
          naturalKeyWidth = _this$props.naturalKeyWidth,
          accidentalWidthRatio = _this$props.accidentalWidthRatio,
          midiNumber = _this$props.midiNumber,
          gliss = _this$props.gliss,
          useTouchEvents = _this$props.useTouchEvents,
          accidental = _this$props.accidental,
          active = _this$props.active,
          disabled = _this$props.disabled,
          children = _this$props.children; // Need to conditionally include/exclude handlers based on useTouchEvents,
    // because otherwise mobile taps double fire events.

    return React.createElement("div", {
      className: classNames('ReactPiano__Key', {
        'ReactPiano__Key--accidental': accidental,
        'ReactPiano__Key--natural': !accidental,
        'ReactPiano__Key--disabled': disabled,
        'ReactPiano__Key--active': active
      }),
      style: {
        left: ratioToPercentage(this.getRelativeKeyPosition(midiNumber) * naturalKeyWidth),
        width: ratioToPercentage(accidental ? accidentalWidthRatio * naturalKeyWidth : naturalKeyWidth)
      },
      onMouseDown: useTouchEvents ? null : this.playNote,
      onMouseUp: useTouchEvents ? null : this.stopNote,
      onMouseEnter: gliss ? this.playNote : null,
      onMouseLeave: this.stopNote,
      onTouchStart: useTouchEvents ? this.playNote : null,
      onTouchCancel: useTouchEvents ? this.stopNote : null,
      onTouchEnd: useTouchEvents ? this.stopNote : null
    }, React.createElement("div", {
      className: "ReactPiano__NoteLabelContainer"
    }, children));
  }

}

_defineProperty(Key, "propTypes", {
  midiNumber: PropTypes.number.isRequired,
  naturalKeyWidth: PropTypes.number.isRequired,
  // Width as a ratio between 0 and 1
  gliss: PropTypes.bool.isRequired,
  useTouchEvents: PropTypes.bool.isRequired,
  accidental: PropTypes.bool.isRequired,
  active: PropTypes.bool.isRequired,
  disabled: PropTypes.bool.isRequired,
  onPlayNote: PropTypes.func.isRequired,
  onStopNote: PropTypes.func.isRequired,
  accidentalWidthRatio: PropTypes.number.isRequired,
  pitchPositions: PropTypes.object.isRequired,
  children: PropTypes.node
});

_defineProperty(Key, "defaultProps", {
  accidentalWidthRatio: 0.65,
  pitchPositions: {
    C: 0,
    Db: 0.55,
    D: 1,
    Eb: 1.8,
    E: 2,
    F: 3,
    Gb: 3.5,
    G: 4,
    Ab: 4.7,
    A: 5,
    Bb: 5.85,
    B: 6
  }
});

function ratioToPercentage(ratio) {
  return `${ratio * 100}%`;
}

class Keyboard extends React.Component {
  // Range of midi numbers on keyboard
  getMidiNumbers() {
    return range(this.props.noteRange.first, this.props.noteRange.last + 1);
  }

  getNaturalKeyCount() {
    return this.getMidiNumbers().filter(number => {
      const _MidiNumbers$getAttri = MidiNumbers.getAttributes(number),
            isAccidental = _MidiNumbers$getAttri.isAccidental;

      return !isAccidental;
    }).length;
  } // Returns a ratio between 0 and 1


  getNaturalKeyWidth() {
    return 1 / this.getNaturalKeyCount();
  }

  getWidth() {
    return this.props.width ? this.props.width : '100%';
  }

  getHeight() {
    if (!this.props.width) {
      return '100%';
    }

    const keyWidth = this.props.width * this.getNaturalKeyWidth();
    return `${keyWidth / this.props.keyWidthToHeight}px`;
  }

  render() {
    const naturalKeyWidth = this.getNaturalKeyWidth();
    return React.createElement("div", {
      className: classNames('ReactPiano__Keyboard', this.props.className),
      style: {
        width: this.getWidth(),
        height: this.getHeight()
      }
    }, this.getMidiNumbers().map(midiNumber => {
      const _MidiNumbers$getAttri2 = MidiNumbers.getAttributes(midiNumber),
            note = _MidiNumbers$getAttri2.note,
            isAccidental = _MidiNumbers$getAttri2.isAccidental;

      const isActive = this.props.activeNotes.includes(midiNumber);
      return React.createElement(Key, {
        naturalKeyWidth: naturalKeyWidth,
        midiNumber: midiNumber,
        noteRange: this.props.noteRange,
        active: isActive,
        accidental: isAccidental,
        disabled: this.props.disabled,
        onPlayNote: this.props.onPlayNote,
        onStopNote: this.props.onStopNote,
        gliss: this.props.gliss,
        useTouchEvents: this.props.useTouchEvents,
        key: midiNumber
      }, this.props.disabled ? null : this.props.renderNoteLabel({
        isActive,
        isAccidental,
        midiNumber
      }));
    }));
  }

}

_defineProperty(Keyboard, "propTypes", {
  noteRange: noteRangePropType,
  activeNotes: PropTypes.arrayOf(PropTypes.number),
  onPlayNote: PropTypes.func.isRequired,
  onStopNote: PropTypes.func.isRequired,
  renderNoteLabel: PropTypes.func.isRequired,
  keyWidthToHeight: PropTypes.number.isRequired,
  className: PropTypes.string,
  disabled: PropTypes.bool,
  gliss: PropTypes.bool,
  useTouchEvents: PropTypes.bool,
  // If width is not provided, must have fixed width and height in parent container
  width: PropTypes.number
});

_defineProperty(Keyboard, "defaultProps", {
  disabled: false,
  gliss: false,
  useTouchEvents: false,
  keyWidthToHeight: 0.25,
  renderNoteLabel: () => {}
});

function isNaturalMidiNumber(value) {
  if (typeof value !== 'number') {
    return false;
  }

  return MidiNumbers.NATURAL_MIDI_NUMBERS.includes(value);
}

function noteRangePropType(props, propName, componentName) {
  const _props$propName = props[propName],
        first = _props$propName.first,
        last = _props$propName.last;

  if (!first || !last) {
    return new Error(`Invalid prop ${propName} supplied to ${componentName}. ${propName} must be an object with .first and .last values.`);
  }

  if (!isNaturalMidiNumber(first) || !isNaturalMidiNumber(last)) {
    return new Error(`Invalid prop ${propName} supplied to ${componentName}. ${propName} values must be valid MIDI numbers, and should not be accidentals (sharp or flat notes).`);
  }

  if (first >= last) {
    return new Error(`Invalid prop ${propName} supplied to ${componentName}. ${propName}.first must be smaller than ${propName}.last.`);
  }
}

class Piano extends React.Component {
  constructor(props) {
    super(props);

    _defineProperty(this, "handleNoteChanges", ({
      prevActiveNotes,
      nextActiveNotes
    }) => {
      const notesStarted = difference(prevActiveNotes, nextActiveNotes);
      const notesStopped = difference(nextActiveNotes, prevActiveNotes);
      notesStarted.forEach(midiNumber => {
        this.onStopNote(midiNumber);
      });
      notesStopped.forEach(midiNumber => {
        this.onPlayNote(midiNumber);
      });
    });

    _defineProperty(this, "getMidiNumberForKey", key => {
      if (!this.props.keyboardShortcuts) {
        return null;
      }

      const shortcut = find(this.props.keyboardShortcuts, {
        key: key
      });
      return shortcut && shortcut.midiNumber;
    });

    _defineProperty(this, "getKeyForMidiNumber", midiNumber => {
      if (!this.props.keyboardShortcuts) {
        return null;
      }

      const shortcut = find(this.props.keyboardShortcuts, {
        midiNumber: midiNumber
      });
      return shortcut && shortcut.key;
    });

    _defineProperty(this, "onKeyDown", event => {
      // Don't conflict with existing combinations like ctrl + t
      if (event.ctrlKey || event.metaKey || event.shiftKey) {
        return;
      }

      const midiNumber = this.getMidiNumberForKey(event.key);

      if (midiNumber) {
        this.onPlayNote(midiNumber);
      }
    });

    _defineProperty(this, "onKeyUp", event => {
      // This *should* also check for event.ctrlKey || event.metaKey || event.ShiftKey like onKeyDown does,
      // but at least on Mac Chrome, when mashing down many alphanumeric keystrokes at once,
      // ctrlKey is fired unexpectedly, which would cause onStopNote to NOT be fired, which causes problematic
      // lingering notes. Since it's fairly safe to call onStopNote even when not necessary,
      // the ctrl/meta/shift check is removed to fix that issue.
      const midiNumber = this.getMidiNumberForKey(event.key);

      if (midiNumber) {
        this.onStopNote(midiNumber);
      }
    });

    _defineProperty(this, "onPlayNote", midiNumber => {
      if (this.props.disabled) {
        return;
      } // Prevent duplicate note firings


      const isActive = this.state.activeNotes.includes(midiNumber);

      if (isActive) {
        return;
      } // Pass in previous activeNotes for recording functionality


      this.props.onPlayNote(midiNumber, this.state.activeNotes);
      this.setState(prevState => ({
        activeNotes: prevState.activeNotes.concat(midiNumber).sort()
      }));
    });

    _defineProperty(this, "onStopNote", midiNumber => {
      if (this.props.disabled) {
        return;
      }

      const isInactive = !this.state.activeNotes.includes(midiNumber);

      if (isInactive) {
        return;
      } // Pass in previous activeNotes for recording functionality


      this.props.onStopNote(midiNumber, this.state.activeNotes);
      this.setState(prevState => ({
        activeNotes: prevState.activeNotes.filter(note => midiNumber !== note)
      }));
    });

    _defineProperty(this, "onMouseDown", () => {
      this.setState({
        isMouseDown: true
      });
    });

    _defineProperty(this, "onMouseUp", () => {
      this.setState({
        isMouseDown: false
      });
    });

    _defineProperty(this, "onTouchStart", () => {
      this.setState({
        useTouchEvents: true
      });
    });

    _defineProperty(this, "renderNoteLabel", ({
      midiNumber,
      isActive,
      isAccidental
    }) => {
      const keyboardShortcut = this.getKeyForMidiNumber(midiNumber);
      return this.props.renderNoteLabel({
        keyboardShortcut,
        midiNumber,
        isActive,
        isAccidental
      });
    });

    this.state = {
      activeNotes: [],
      isMouseDown: false,
      useTouchEvents: false
    };
  }

  componentDidMount() {
    window.addEventListener('keydown', this.onKeyDown);
    window.addEventListener('keyup', this.onKeyUp);
    window.addEventListener('mousedown', this.onMouseDown);
    window.addEventListener('mouseup', this.onMouseUp);
    window.addEventListener('touchstart', this.onTouchStart);
  }

  componentWillUnmount() {
    window.removeEventListener('keydown', this.onKeyDown);
    window.removeEventListener('keyup', this.onKeyUp);
    window.removeEventListener('mousedown', this.onMouseDown);
    window.removeEventListener('mouseup', this.onMouseUp);
    window.removeEventListener('touchstart', this.onTouchStart);
  }

  componentDidUpdate(prevProps, prevState) {
    // If keyboard shortcuts change, any activeNotes will not
    // correctly fire keyUp events. First cancel all activeNotes.
    if (prevProps.keyboardShortcuts !== this.props.keyboardShortcuts) {
      this.setState({
        activeNotes: []
      });
    }

    if (this.state.activeNotes !== prevState.activeNotes) {
      this.handleNoteChanges({
        prevActiveNotes: prevState.activeNotes || [],
        nextActiveNotes: this.state.activeNotes || []
      });
    }
  }

  render() {
    return React.createElement(Keyboard, {
      noteRange: this.props.noteRange,
      onPlayNote: this.onPlayNote,
      onStopNote: this.onStopNote,
      activeNotes: this.props.playbackNotes || this.state.activeNotes,
      className: this.props.className,
      disabled: this.props.disabled,
      width: this.props.width,
      keyWidthToHeight: this.props.keyWidthToHeight,
      gliss: this.state.isMouseDown,
      useTouchEvents: this.state.useTouchEvents,
      renderNoteLabel: this.renderNoteLabel
    });
  }

}

_defineProperty(Piano, "propTypes", {
  noteRange: PropTypes.object.isRequired,
  onPlayNote: PropTypes.func.isRequired,
  onStopNote: PropTypes.func.isRequired,
  renderNoteLabel: PropTypes.func.isRequired,
  className: PropTypes.string,
  disabled: PropTypes.bool,
  width: PropTypes.number,
  keyWidthToHeight: PropTypes.number,
  keyboardShortcuts: PropTypes.arrayOf(PropTypes.shape({
    key: PropTypes.string.isRequired,
    midiNumber: PropTypes.number.isRequired
  })),
  playbackNotes: PropTypes.arrayOf(PropTypes.number.isRequired)
});

_defineProperty(Piano, "defaultProps", {
  renderNoteLabel: ({
    keyboardShortcut,
    midiNumber,
    isActive,
    isAccidental
  }) => keyboardShortcut ? React.createElement("div", {
    className: classNames('ReactPiano__NoteLabel', {
      'ReactPiano__NoteLabel--active': isActive,
      'ReactPiano__NoteLabel--accidental': isAccidental,
      'ReactPiano__NoteLabel--natural': !isAccidental
    })
  }, keyboardShortcut) : null
});

function createKeyboardShortcuts({
  firstNote,
  lastNote,
  keyboardConfig
}) {
  let currentMidiNumber = firstNote;
  let naturalKeyIndex = 0;
  let keyboardShortcuts = [];

  while ( // There are still keys to be assigned
  naturalKeyIndex < keyboardConfig.length && // Note to be assigned does not surpass range
  currentMidiNumber <= lastNote) {
    const key = keyboardConfig[naturalKeyIndex];

    const _MidiNumbers$getAttri = MidiNumbers.getAttributes(currentMidiNumber),
          isAccidental = _MidiNumbers$getAttri.isAccidental;

    if (isAccidental) {
      keyboardShortcuts.push({
        key: key.flat,
        midiNumber: currentMidiNumber
      });
    } else {
      keyboardShortcuts.push({
        key: key.natural,
        midiNumber: currentMidiNumber
      });
      naturalKeyIndex += 1;
    }

    currentMidiNumber += 1;
  }

  return keyboardShortcuts;
}

var KeyboardShortcuts = {
  create: createKeyboardShortcuts,
  // Preset configurations
  BOTTOM_ROW: [{
    natural: 'z',
    flat: 'a',
    sharp: 's'
  }, {
    natural: 'x',
    flat: 's',
    sharp: 'd'
  }, {
    natural: 'c',
    flat: 'd',
    sharp: 'f'
  }, {
    natural: 'v',
    flat: 'f',
    sharp: 'g'
  }, {
    natural: 'b',
    flat: 'g',
    sharp: 'h'
  }, {
    natural: 'n',
    flat: 'h',
    sharp: 'j'
  }, {
    natural: 'm',
    flat: 'j',
    sharp: 'k'
  }, {
    natural: ',',
    flat: 'k',
    sharp: 'l'
  }, {
    natural: '.',
    flat: 'l',
    sharp: ';'
  }, {
    natural: '/',
    flat: ';',
    sharp: "'"
  }],
  HOME_ROW: [{
    natural: 'a',
    flat: 'q',
    sharp: 'w'
  }, {
    natural: 's',
    flat: 'w',
    sharp: 'e'
  }, {
    natural: 'd',
    flat: 'e',
    sharp: 'r'
  }, {
    natural: 'f',
    flat: 'r',
    sharp: 't'
  }, {
    natural: 'g',
    flat: 't',
    sharp: 'y'
  }, {
    natural: 'h',
    flat: 'y',
    sharp: 'u'
  }, {
    natural: 'j',
    flat: 'u',
    sharp: 'i'
  }, {
    natural: 'k',
    flat: 'i',
    sharp: 'o'
  }, {
    natural: 'l',
    flat: 'o',
    sharp: 'p'
  }, {
    natural: ';',
    flat: 'p',
    sharp: '['
  }, {
    natural: "'",
    flat: '[',
    sharp: ']'
  }],
  QWERTY_ROW: [{
    natural: 'q',
    flat: '1',
    sharp: '2'
  }, {
    natural: 'w',
    flat: '2',
    sharp: '3'
  }, {
    natural: 'e',
    flat: '3',
    sharp: '4'
  }, {
    natural: 'r',
    flat: '4',
    sharp: '5'
  }, {
    natural: 't',
    flat: '5',
    sharp: '6'
  }, {
    natural: 'y',
    flat: '6',
    sharp: '7'
  }, {
    natural: 'u',
    flat: '7',
    sharp: '8'
  }, {
    natural: 'i',
    flat: '8',
    sharp: '9'
  }, {
    natural: 'o',
    flat: '9',
    sharp: '0'
  }, {
    natural: 'p',
    flat: '0',
    sharp: '-'
  }, {
    natural: '[',
    flat: '-',
    sharp: '='
  }]
};

export { Piano, Keyboard, KeyboardShortcuts, MidiNumbers };
//# sourceMappingURL=react-piano.esm.js.map
